# Copyright 2007 by Tiago Antao.  All rights reserved.
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.


from Bio.PopGen import GenePop

"""
Utility functions to deal with GenePop files

This had some functions that where moved to
GenePop.Record (now methods of that class).

This file is maintained as it is believed it will have
utility functions in the future again.
"""
